import re
from robot.utils import normalize, unic


def apply_all(libdoc):
    """Apply all patches."""
    patch_scope(libdoc)
    patch_params(libdoc)


def patch_scope(libdoc):
    """Convert documentation scope to RPA terminology."""
    normalized = normalize(unic(libdoc.scope), ignore="_")
    libdoc.scope = {"testcase": "Task", "testsuite": "Suite", "global": "Global"}.get(
        normalized, libdoc.scope
    )


def patch_params(libdoc):
    """Patch malformed params tables in reST documentation."""
    if libdoc.doc_format.strip().upper() != "REST":
        return

    def _replace_all(text):
        return re.sub(r":param\s+(.*):\s+", r":\1: ", text)

    try:
        libdoc.doc = _replace_all(libdoc.doc)
    except AttributeError:
        libdoc._doc = _replace_all(libdoc.doc)

    for init in libdoc.inits:
        init.doc = _replace_all(init.doc)
    for keyword in libdoc.keywords:
        keyword.doc = _replace_all(keyword.doc)



import pkgutil
import os
import re
import sys
from functools import lru_cache
from pathlib import Path
from robot.libraries import STDLIBS
from docgen import loader, utils

RESOURCES = (".robot", ".resource", ".txt")
INIT_FILES = ("__init__.py", "__init__.robot", "__init__.txt")

KEYWORDS = re.compile(r"^\*+\s*((?:User )?Keywords?)", re.MULTILINE | re.IGNORECASE)
TASKS = re.compile(r"^\*+\s*(Test Cases?|Tasks?)", re.MULTILINE | re.IGNORECASE)


def find_all(curdir=True):
    matches = set()

    matches.update(builtins())
    matches.update(rpaframework())
    matches.update(walk_modules())

    if curdir:
        matches.update(walk_directory(os.getcwd()))

    return [str(match) for match in matches]


def builtins():
    return [name for name in STDLIBS if name not in ("Reserved", "Remote")]


def rpaframework():
    paths = set()
    for path in sys.path:
        namespace = Path(path) / "RPA"
        if namespace.is_dir():
            paths.update(walk_directory(namespace))
    return ["RPA." + utils.path_to_name(path) for path in paths]


def walk_modules(root=None):
    modules = []
    for module in pkgutil.iter_modules(root):
        if module.name.startswith("_"):
            continue
        elif module.ispkg:
            root = (Path(module.module_finder.path) / module.name).resolve()
            modules.extend(module.name + "." + name for name in walk_modules([root]))
            modules.append(module.name)
        else:
            modules.append(module.name)
    return modules


def walk_directory(root, libraries=True, resources=True):
    paths, stack = set(), [Path(root)]
    while stack:
        path = stack.pop(0)
        path = path.resolve()

        if path.name.startswith(("_", ".")) or path.parts[-1] == "__pycache__":
            continue
        try:
            if path.is_dir():
                if libraries and is_module_library(path):
                    paths.add(path)
                for child in path.iterdir():
                    stack.append(child)
            elif libraries and is_library_file(path):
                paths.add(path)
            elif resources and is_resource_file(path):
                paths.add(path)
        except Exception:
            utils.debug_traceback()

    return [path.relative_to(root) for path in paths]


def is_module_library(path):
    if not any((path / init).is_file() for init in INIT_FILES):
        return False

    try:
        return bool(loader.load(path).keywords)
    except ImportError:
        return False


def is_library_file(path):
    return path.suffix == ".py" and path.name != "__init__.py"


@lru_cache(maxsize=None)
def is_resource_file(path):
    if path.name in INIT_FILES or path.suffix not in RESOURCES:
        return False

    with open(path, "r", encoding="utf-8", errors="ignore") as infile:
        content = infile.read()

    has_keywords = bool(re.search(KEYWORDS, content))
    has_tasks = bool(re.search(TASKS, content))

    return not has_tasks and has_keywords



import logging
import os
import time
import traceback
from contextlib import contextmanager
from pathlib import Path


@contextmanager
def timed(name="Task"):
    """Timer block for long-running operations."""
    start = time.time()
    try:
        yield
    finally:
        duration = time.time() - start
        logging.info("Task '%s' completed in %.3f seconds", name, duration)


@contextmanager
def silent():
    """Suppress messages from logging module within block."""
    logger = logging.getLogger()
    previous = logger.getEffectiveLevel()
    try:
        logger.setLevel(logging.CRITICAL + 1)
        yield
    finally:
        logger.setLevel(previous)


def debug_traceback():
    """Log traceback if DEBUG level is enabled."""
    if logging.getLogger().isEnabledFor(logging.DEBUG):
        traceback.print_exc()


def path_to_name(path):
    """Convert relative path to Robot Framework import name."""
    if path.parent != Path("."):
        namespace = str(path.parent).replace(os.sep, ".")
        return "{}.{}".format(namespace, path.stem)
    else:
        return path.stem



import abc
import json
import logging
import re
from collections import defaultdict
from pathlib import Path
from robot.libdocpkg import htmlwriter
from robot.libdocpkg.model import LibraryDoc
from docgen import rest
from robot.version import VERSION

RF_VERSION = VERSION

# Populated dynamically
CONVERTERS = {}


def convert(libdoc: LibraryDoc, fmt, dirname):
    """Convert libdoc model into output file."""
    converter = CONVERTERS.get(fmt)
    if not converter:
        raise ValueError(f"Unknown conversion format: {fmt}")

    filename = Path(libdoc.name.replace(".", "_"))
    filename = filename.with_suffix(converter.EXTENSION)

    path = str(Path(dirname) / filename)
    converter().convert(libdoc, path)
    logging.info("Created file: %s", path)


class ConverterMeta(abc.ABCMeta):
    def __new__(cls, name, bases, namespace, **kwargs):
        converter = super().__new__(cls, name, bases, namespace, **kwargs)
        if name == "BaseConverter":
            return converter

        if not getattr(converter, "NAME", None):
            raise ValueError(f"Undefined 'NAME' attribute in {converter}")
        if not getattr(converter, "EXTENSION", None):
            raise ValueError(f"Undefined 'EXTENSION' attribute in {converter}")
        if converter.NAME in CONVERTERS:
            raise ValueError(f"Duplicate converter for {converter.NAME}")

        CONVERTERS[converter.NAME] = converter
        return converter


class BaseConverter(metaclass=ConverterMeta):
    NAME = None
    EXTENSION = None

    @abc.abstractmethod
    def convert(self, libdoc: LibraryDoc, output):
        raise NotImplementedError


class JsonConverter(BaseConverter):
    NAME = "json"
    EXTENSION = ".json"

    class _NullFormatter:
        def html(self, doc, *args, **kwargs):
            return doc

    def convert(self, libdoc: LibraryDoc, output):
        if int(RF_VERSION[0]) <= 3:
            data = htmlwriter.JsonConverter(self._NullFormatter()).convert(libdoc)
            with open(output, "w") as fd:
                json.dump(data, fd, indent=4)
        else:
            libdoc.save(output, "JSON")


class JsonHtmlConverter(BaseConverter):
    NAME = "json-html"
    EXTENSION = ".json"

    def convert(self, libdoc: LibraryDoc, output):
        if int(RF_VERSION[0]) <= 3:
            formatter = htmlwriter.DocFormatter(
                libdoc.keywords, libdoc.doc, libdoc.doc_format
            )
            data = htmlwriter.JsonConverter(formatter).convert(libdoc)

            # Format keyword shortdoc with correct formatter, strip extra HTML tags
            for keyword in data["keywords"]:
                shortdoc = formatter.html(keyword["shortdoc"])
                shortdoc = re.search(r"<p>(.*)<\/p>", shortdoc)

                if shortdoc is None:
                    logging.warning(
                        "Missing shortdoc: %s.%s", data["name"], keyword["name"]
                    )
                    continue

                keyword["shortdoc"] = shortdoc.group(1)

            with open(output, "w") as fd:
                json.dump(data, fd, indent=4)
        else:
            libdoc.convert_docs_to_html()
            libdoc.save(output, "JSON")


class XmlConverter(BaseConverter):
    NAME = "xml"
    EXTENSION = ".xml"

    def convert(self, libdoc: LibraryDoc, output):
        libdoc.save(output, "XML")


class HtmlConverter(BaseConverter):
    NAME = "html"
    EXTENSION = ".html"

    def convert(self, libdoc: LibraryDoc, output):
        if int(RF_VERSION[0]) >= 4:
            libdoc.convert_docs_to_html()
        libdoc.save(output, "HTML")


class XmlHtmlConverter(BaseConverter):
    NAME = "xml-html"
    EXTENSION = ".html"

    def convert(self, libdoc: LibraryDoc, output):
        if int(RF_VERSION[0]) <= 3:
            libdoc.save(output, "XML:HTML")
        else:
            libdoc.convert_docs_to_html()
            libdoc.save(output, "XML")


class RestConverter(BaseConverter):
    NAME = "rest"
    EXTENSION = ".rst"

    IGNORE = (r"^:param.*", r"^:return.*")

    def __init__(self):
        self.ignore_block = False

    def convert(self, libdoc: LibraryDoc, output):
        writer = rest.RestWriter()
        with writer.heading(libdoc.name):
            self.overview(writer, libdoc)
            self.inits(writer, libdoc)
            self.keywords(writer, libdoc)
        writer.save(output)

    def filter_docstring(self, text):
        output = []
        for line in text.split("\n"):
            if any(re.match(pattern, line) for pattern in self.IGNORE):
                self.ignore_block = True
                continue
            if line.startswith(" ") and self.ignore_block:
                continue

            self.ignore_block = False
            output.append(line)
        return "\n".join(output)

    @staticmethod
    def escape_string(text: str):
        return text.replace("*", "\\*")

    def overview(self, writer, libdoc: LibraryDoc):
        with writer.heading("Description"):
            writer.fieldlist(("Library scope", libdoc.scope))
            writer.raw(self.filter_docstring(libdoc.doc))

    def inits(self, writer, libdoc: LibraryDoc):
        with writer.heading("Init"):
            for init in libdoc.inits:
                writer.raw(init.doc)

    def keywords(self, writer, libdoc: LibraryDoc):
        groups = defaultdict(list)
        for keyword in libdoc.keywords:
            name = Path(keyword.source).stem
            groups[name].append(keyword)

        def _init_first(string):
            return string == "__init__", string

        with writer.heading("Keywords"):
            if len(groups) > 1:
                for group, keywords in sorted(groups.items(), key=_init_first):
                    group = "main" if group == "__init__" else group
                    with writer.heading(group.title()):
                        for keyword in keywords:
                            self._keyword(writer, keyword)
            else:
                keywords = next(iter(groups.values())) if groups else []
                for keyword in keywords:
                    self._keyword(writer, keyword)

    def _keyword(self, writer, keyword):
        with writer.field(keyword.name):
            fields = []
            if keyword.args:
                if int(RF_VERSION[0]) <= 3:
                    args = (self.escape_string(arg) for arg in keyword.args)
                    fields.append(("Arguments", ", ".join(args)))
                else:

                    args = (str(arg) for arg in keyword.args)
                    fields.append(("Arguments", ", ".join(args)))

            if keyword.tags:
                fields.append(("Tags", ", ".join(keyword.tags)))
            writer.fieldlist(*fields)
            writer.raw(self.filter_docstring(keyword.doc))


class RestHtmlConverter(RestConverter):
    NAME = "rest-html"
    EXTENSION = ".rst"

    def convert(self, libdoc: LibraryDoc, output):
        if int(RF_VERSION[0]) <= 3:
            DocFormatter = htmlwriter.DocFormatter
            formatter = DocFormatter(libdoc.keywords, libdoc.doc, libdoc.doc_format)
        else:
            from robot.libdocpkg.htmlutils import DocFormatter

            formatter = DocFormatter(
                libdoc.keywords, libdoc.data_types, libdoc.doc, libdoc.doc_format
            )

        doc = self._raw_html(formatter.html(libdoc.doc))
        try:
            # Robot Framework < 3.2
            libdoc.doc = doc
        except AttributeError:
            # Robot Framework >= 3.2
            libdoc._doc = doc
        for init in libdoc.inits:
            init.doc = self._raw_html(formatter.html(init.doc))
        for kw in libdoc.keywords:
            kw.doc = self._raw_html(formatter.html(kw.doc))
        if int(RF_VERSION[0]) >= 4:
            for type in libdoc.data_types:
                type.doc = self._raw_html(formatter.html(type.doc))

        super().convert(libdoc, output)

    def _raw_html(self, content):
        output = [".. raw:: html", ""]
        for line in content.splitlines():
            output.append(f"   {line}")
        return "\n".join(output)



from contextlib import contextmanager


class RestWriter:
    """Helper class for writing reStructuredText"""

    INDENT = "  "

    def __init__(self):
        self.body = []
        self._indent = 0
        self._section = 0

    def save(self, path):
        with open(path, "w") as fd:
            fd.write(self.as_text())

    def write(self, text=""):
        for line in text.split("\n"):
            # Do not indent empty lines
            if not line.strip():
                self.body.append("")
                continue

            self.body.append(
                "{indent}{line}".format(indent=self.INDENT * self._indent, line=line)
            )

    def as_text(self):
        return "\n".join(self.body)

    def raw(self, text):
        self.write(text)
        self.write()

    @contextmanager
    def heading(self, content):
        try:
            self._heading(content)
            self._section += 1
            yield
        finally:
            self._section -= 1

    def _heading(self, content):
        chars = "#*=-"
        line = chars[self._section] * len(content)

        if self._section < 2:
            self.write(line)
        self.write(content)
        self.write(line)
        self.write()

    @contextmanager
    def field(self, name):
        try:
            self.write(f":{name}:")
            self._indent += 1
            yield
        finally:
            self._indent -= 1

    def fieldlist(self, *values):
        if not values:
            return
        for name, value in values:
            self.write(f":{name}: {value}")
        self.write()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['docgen']

package_data = \
{'': ['*']}

install_requires = \
['robotframework>=3.2.2,<5']

entry_points = \
{'console_scripts': ['docgen = docgen.cli:run']}

setup_kwargs = {
    'name': 'robotframework-docgen',
    'version': '0.8.1',
    'description': 'Robot Framework documentation finder and generator',
    'long_description': '',
    'author': 'Ossi Rajuvaara',
    'author_email': 'ossi@robocorp.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/robocorp/robotframework-docgen',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
"""Robot Framework documentation finder and generator.

Example usage:
> docgen -f html Process
> docgen RPA.*

"""
import argparse
import fnmatch
import logging
import multiprocessing
import platform
import sys
from pathlib import Path

from docgen import finder, loader, patcher, converter
from docgen.utils import timed, debug_traceback


DEFAULT_EXCLUDE = [
    r"_*",
    r"*.setup",
    r"*.main",
    r"antigravity",
    r"robot.libraries.Remote",
    r"robot.libraries.Reserved",
]


def filter_include(matches, patterns):
    output = []
    for match in matches:
        for pattern in patterns:
            if fnmatch.fnmatch(match, pattern):
                output.append(match)
                break
    return output


def filter_exclude(matches, patterns):
    output = []
    for match in matches:
        for pattern in patterns:
            if fnmatch.fnmatch(match, pattern):
                logging.debug("Ignoring %s", match)
                break
        else:
            output.append(match)
    return output


def create_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "pattern",
        help="library name pattern (default: %(default)s)",
        nargs="?",
        default="*",
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output directory (default: %(default)s)",
        type=Path,
        default="dist",
    )
    parser.add_argument(
        "-f",
        "--format",
        help="output format (default: %(default)s)",
        choices=converter.CONVERTERS,
        default="json-html",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        help="pattern to exclude",
        action="append",
        default=DEFAULT_EXCLUDE,
    )
    parser.add_argument(
        "--no-patches",
        action="store_true",
        help="Don't apply automatic patches to generated output",
    )
    parser.add_argument(
        "--no-curdir",
        action="store_true",
        help="Don't parse resources from current directory",
    )
    parser.add_argument(
        "-v", "--verbose", help="be more talkative", action="store_true"
    )
    return parser


def find(args):
    if args.pattern == "robotframework":
        matches = finder.builtins()
        include = "*"
    elif args.pattern == "rpaframework":
        matches = finder.rpaframework()
        include = "*"
    else:
        matches = finder.find_all(curdir=(not args.no_curdir))
        include = args.pattern

    return matches, include


def try_load(match):
    try:
        return loader.load(match)
    except ImportError:
        debug_traceback()


def run(raw_args=None):
    parser = create_parser()
    args = parser.parse_args(raw_args)

    level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(asctime)s %(levelname)-8s %(message)s",
    )

    if platform.system() == "Darwin":
        multiprocessing.set_start_method("spawn")

    with multiprocessing.Pool() as pool:
        with timed("All"):
            with timed("Find"):
                result = pool.map(find, [args])
                matches, include = result[0]

            with timed("Filter"):
                matches = filter_include(matches, [include])
                matches = filter_exclude(matches, args.exclude)

            with timed("Load"):
                libraries = pool.map(try_load, matches)
                libraries = [lib for lib in libraries if lib is not None]

            if not args.no_patches:
                with timed("Patch"):
                    for lib in libraries:
                        patcher.apply_all(lib)

            args.output.mkdir(parents=True, exist_ok=True)
            with timed("Convert"):
                for lib in libraries:
                    converter.convert(lib, args.format, args.output)

        logging.info("Created %d file(s)", len(libraries))



from robot.libdocpkg import LibraryDocumentation
from docgen import utils


def load(value):
    """Parse library into container, with caching."""
    try:
        with utils.silent():
            return LibraryDocumentation(str(value))
    except (Exception, SystemExit):
        utils.debug_traceback()
        raise ImportError(f"Failed to load library: {value}")
